BOOK_TOKEN_POLICY_ID = "51a5e236c4de3af2b8020442e2a26f454fda3b04cb621c1294a0ef34"
BOOK_TOKEN_ASSET_NAME = "424f4f4b"
BOOK_TOKEN_ASSET_ID = f"{BOOK_TOKEN_POLICY_ID}{BOOK_TOKEN_ASSET_NAME}"
BOOK_TOKEN_FINGERPRINT = "asset1lpmhzq33vsj9jpj2wsluu0usqfaf6kk96zdw5c"
BOOK_TOKEN_SYMBOL = "BOOK"
